# neurolab-mongo-python

![image](https://user-images.githubusercontent.com/57321948/196933065-4b16c235-f3b9-4391-9cfe-4affcec87c35.png)

### Step 1 - Install the requirements

```bash
pip install -r requirements.txt
```

### Step 2 - Run main.py file

```bash
python main.py
```

* Ineuron Lab 所建立的 MongoDB 範本，但無法直接使用 ineuron lab ，此 section 是手動建立的（從sunnysavita10/MongoDB-Connector-PYPI-Package 的初始 commit 內容複製而來）！

* 從 sunnysavita10/MongoDB-Connector-PYPI-Package 所在 github 執行 fork，後用 vscode.dev 開啟自己的 repository，然後將一些檔案刪除到如（23:00）這般，再 push 到自己的 repository，完成類似用 ineuron lab 所建立的 MongoDB 範本，接著用 Gitpod.io 開啟自己的 Github repository，接著就可以按照 Lecture 的步驟進行操作!

---
# Lecture 1 Note -- [MLOps Foundations and Fundamentals](https://www.youtube.com/watch?v=jpU8F0M5axo)

* (21:45) The prerequisite: The basics of the machine learning
  > 1. Basics of ML/DL
  > 
  > 2. Basics of the Python

## (29:30) MLOps -- DevOps: SDLC(software development live cycle)

* 5 phases in Software Development live cycle (五階段)
  > 1. Planning & Requirement
  > 
  > 2. Design
  > 
  > 3. Building -- Dev. w.r.t any sort of a language like python, Java, ... 
  > 
  > 4. Testing
  > 
  > 5. Deploy

* Types of software development models
  > 1. Waterfall Model
  > 
  > 2. Agile methodology (state-of-art) – deliver a APP in a shortest time with a minimalistic feature

## (45:20) Corona scenario

* 開發 Aarogya Setu 應用程式 -- (顧客 Government)，(賣方 Organization)有以下兩者：
  > 交給 XYZ 公司 ：需 8~10 months，遵循 SDLC 五階段 \
  > 交給 abc 公司 ：需 2~3 weeks，每一個iteration 只實現簡單一部份功能！

* launch the APP in iterations. At the first place, create a simple UI. Take the information from the user. Based on that, show the basic guidelines/ only the basic detail. In the next iteration, keep it to the advanced one. Add a few more information(e.g., regarding traveling and all). In the third iteration, again add something over there. So they are working in iteration. The first iteration follows all the step regrading 五階段 in waterfall + production. Deploy and then, the APP is visible to everyone. In the second iteration, again, they came up, again, they have planned with a new feature and they have implemented that and then basically they have tested and then deployed. So, they are working on iteration-wise. They are not going to build up the entire APP in a single shot and then they are going to test it and then they are going to be deployed.

* So the waterfall, it’s having many disadvantages. It's not for the complex project and it's not a Time effective. On the other hand, agile, it is working in iterations. Therefore, nowadays in a market actually we are using agile, not a waterfall because agile is the state-of-art and with that basically we are able to deliver an APP in a shortest time and with a minimalistic feature also.

## (53:00) DevOps

* Dev+Ops
  > * Development
  > |Dev|: Java/Python/PHP, Git/GitHub \n
  > --> |testing| \n
  > --> |QA| Jenkins/Github Action/Circle ci/Travis ci
  >
  > * Operations
  > --> |Delivery|: Docker/Kubernetes \n
  > --> |Deployment| \n
  > --> |maintenance/monitoring|

* Manual (every step in Waterfall) VS. Automated (the entire process)

* two methodologies (whatever application we are building nowadays)
  > Agile
  >
  > DevOps

## (1:06:23) MLOps
* Development:
|Buidling| 
--> |Testing|: manual/automated/UAT(user acceptance testing)/unit/integration 
-----> |QA(quality assurance)|

* Operation:
|Delivery|
-->|Deployment|
---->|monitor/maintenance|

* MLOps: ML + Ops

  > ML: 
  > 1. Data
  > 2. Validation
  > 3. EDA--Exploratory Data Analysis
  > 4. Feature Engineering
  > 5. Model building/ Model Evaluation

## (1:13:44) ML & DL 

* Machine Learning -- Regression, Classification
* Deep Learning -- Vision, NLP => App
  > <pre>採用Agile -- |MLOPS| : building a ML APP & responsible for maintaining the entire pipeline. \
  >                 ^ 
  >                 ^ 
  > 採用Agile -- |DEVOPS| 
  >                 ^ 
  >                 ^ 
  >              |SDLC|</pre>

## (1:21:55) Flow regarding one application (MLOps engineers)
* ML-APP
  > <pre>|Build|: Python, Git..(tools) ~ [local] 
  > -->|test|: unit, integration ~ [central repo/github]
  > -->|Deliver|: Docker image ~ [Docker Hub/Amazon]
  > -->|Deploy|: container ~  [AWS]
  > -->|Monitor/Maintain| </pre>

* (1:30:40) More terms in market
  > <pre>Amazon - DevOps、MLOps
  > GCP - DevOps、MLOps
  > Azure - DevOps、MLOps</pre>

* open source tools (for MLOps engineers):
  > Git、CircleCI、Terraform、Ansible、Docker Hub

## (1:33:55) CI/CD/CD + CT (Continuous Training): Air flow

* CI
  > |Develop| (local) 

  > --> |testing|

  > --> |QA|

* CD
  > --> |Deliver|

* CD
  > --> |Deploy|

  > --> |Monitor/Maintain|

* (1:38:07) MLOPS engineers 的工作
  > CI/CD/CD/CT

* (1:43:30) MLOps Architecture [MLOps: Continuous delivery and automation pipelines in machine learning](https://cloud.google.com/architecture/mlops-continuous-delivery-and-automation-pipelines-in-machine-learning)

* (1:52:00) Create a GitHub Repository 

# Lecture 2 Note -- [MLOps End-to-End Project Setup with GitHub](https://www.youtube.com/watch?v=6qnx3okPvkc)
* MLOPS(ML- + [OPS]):
  > <pre>1-. Data Ingestion/collection 
  >     > Data Pipeline: Source to Destination - Data engineering
  > 2-. EDA
  > 3-. FE--feature engineering
  > 4-. Model Creation
  > 5-. Model Evaluation
  > [6]. Deploy
  > [7]. Monitor
  > [8]. Retrain </pre>

* (37:55)Devops VS. MLOps

## (48:00) 建立 `template.py`

```python
import os
from pathlib import Path

package_name = "mongodb_connect"

list_of_files = [
   ".github/workflows/.gitkeep",
   "src/components/__init__.py",
   "src/components/data_ingestion.py",
"src/components/data_transformation.py",
"src/components/data_trainer.py",
"src/components/data_evaluation.py",
"src/pipeline/__init__.py",
"src/pipeline/training_pipeline.py",
"src/pipeline/prediction_pipeline.py",
"src/utils/__init__.py",
"src/utils/utils.py",
"src/logger/logging.py",
"src/exceptions/exception.py",
"tests/utils/__init__.py",
"tests/integration/__init__.py",
"init_setup.sh",
"requirements.txt",
"requirements_dev.txt",
"setup.py",
"setup.cfg",
"pyproject.toml",
"tox.ini",
"experiment/experiments.ipynb"
]

for filepath in list_of_files:
    filepath = Path(filepath)
    filedir, filename = os.path.split(filepath)
    if filedir != "":
        os.makedirs(filedir, exist_ok=True)
        logging.info(f"Creating directory: {filedir} for file: {filename}")

    if (not os.path.exists(filepath)) or (os.path.getsize(filepath) == 0):
        with open(filepath, "w") as f:
            pass # create an empty file
```

* 執行 `python template.py`

* 執行 git push
  > ```bash
  > git add .
  > git commit -m "folder structure updated" # 出錯
  > 
  > git config --global user.email "u860218@gmail.com"
  > git config --global user.name "henrykohl"
  > git commit -m "folder structure updated" # 再次執行
  > git push -f origin main
  > ```
## Machine Learning:

* Supervised VS. Unsupervised

* Two stages
  > 1. (tp): training -- pipeline
  > 
  > 2. (vp): validation/testing/inferencing -- pipeline

* (tp):
  > |Data| -> |Data Validation| -> |FE| -> |Model training| -> |Model evaluation|

* (vp):
  > <pre>|Data| -> |Data Validation| -> |FE| -> |Model evaluation|
  > Two types of testing
  > Single value prediction
  > Bulkd prediciton<pre>

* 執行 git push
  > ```bash
  > git add .
  > git commit -m "structure_updated"
  > git push -f origin main
  > ```

## (1:48:00) Discuss something before the main project
<pre>(1) Introduciton
(2) Structure
(3) Python Package -> PyPi</pre>

* CI
  > github action
  >
  > github page

* MongoDB -- Database

* `setup.py`、`setup.cgf`、`pyproject.toml`、`tox.ini` are required

* (1:55:15) Demo GitHub Template repositories 

# Lecture 3 Note -- [How to Create Python Package for MLOps Project](https://www.youtube.com/watch?v=vKi-l__1xg0)

* GitHub Repository Resource 1 -- [MongoDB-Connector-PYPI-Package](https://github.com/sunnysavita10/MongoDB-Connector-PYPI-Package)

* GitHub Repository Resource 2 -- [MongoDB-Connector](https://github.com/sunnysavita10/MongoDB-Connector)

* (34:15) 執行 `python template.py`

## Problem statement (38:50)

* Python Package
       |
       ----> folder
                |
                ----> multiple files (python)

* Problem statement: Release our own package -> Push our entire code to the Github

* In this lecture, we create our own package.

* (46:30) MongoDB --> NoSql => Document{Key: value} (Dict, JSON, ...)
           |
           -----> Local server
           |
           -----> Cloud service (Atlas)  
  > If using Python-Client, we'll have to write down some sort of python scripts

* Unified package -- Unified Python package
  > write down the code (follow the documentation guideline): 1. MongoDB 2. Cassandra 3. My-sql
  >
  > --> single function -- passing different parameters: 1. URL 2. DB name 3. Collection -- (e.g., `insert_data()` to store a single row or multiple rows). So in such a way, we  can customize my code and publish the code in terms of the package. 
  >
  > We are going to create one python package. The python package name is a DB connector or MongoDB connector.  Initially, we will write down the code for MongoDB inside our own customized package. We will customize all the steps whatever step is required for connecting with MongoDB. Customizing in such a way, we don't need to follow all the step again and again. We create our package. Inside the package, it means that inside the folder we will be having a python file. We will write down the MongoDB code in such a way. You can download the package. Then, you can utilize those thing in a unified manner. No need to follow all the step again and again (e.g., database creation, collection creation and all).


## Code (54:50)

* 完成 `requirements.txt`、`requirements_dev.txt`、`setup.py`、`pyproject.toml`、`tox.ini`，執行 `git push`
  > ```bash
  > git init
  > git add .
  > git commit -m "code updated"
  > git config --global user.email "github郵箱地址"
  > git config --global user.name "github帳號名稱"
  > git branch -M main # 此步驟沒執行，git push 會出錯
  > git remote add origin https://github.com/henrykohl
  > git push -u origin main
  > ```

* code explanation
  > (1:10:45) `requirements_dev.txt`: requirement for development environment. dsnpython -- for MongoDB. ensure -- use for ensuring something. pytest -- test for use cases. tox (用在tox.ini) --  using the tox, we can create a testing environment. black and flask8 -- linting tools. \n
  > (1:14:55) `setup.py`: 在 Lecture 3 中加入 `install_requires=get_requirement("requirements_dev.txt")`，自行實現時，在 GitHub Action 中執行 Test with tox 時，找不到 `"requirements_dev.txt"`，不知道為何！後來才google到解決方式~建立 `MANIFEST.in` 加入 `include requirements_dev.txt`
  > (1:23:45) `setup.cfg`: 檔案中 每一個 tag/title - 也就是 [] - 後面是 information ，有對應的 Key & value. It's just an extension of `setup.py`, representing a configuration of `setup.py`. \
  > (1:30:40) `tox.ini` (testing for development environment) are using for testing our code in local environment in a development environment if we want our code in a local environment or while we are going to integrate it. So, while performing this CI, I'm going to perform it by using the GitHub. GitHub is also having one service which provides us a server -- the service name is called GitHub Action. Once it creats the python environment, the first place it will create a python environment. Then, it will install the dependency inside that environment. In [testenv], `deps=...` will install the requirements_dev.txt. Finally, it's going to perform the following particular commands (`commands=...`). `flake8 src ...` means that going into the `src` folder, it is using the `flake8`. `mypy src/` means that checking `src/` we are using `mypy`. `mypy` is a linting tool, which is able to check whether the code is correct. Using `flake8` for all our code is following all the protocol or not. Again, we are going to do testing by using the tox.ini file. `pytest -v ...` will do unit testing and integration testing.
  > Provide you the local environment for testing your application (the number of environments can be more than one). We can create the environments for different versions. PS. `-v` is verbose -- whatever execution is happening in backend, you are going to see all the execution in your screen. PS. `--count` is the command line argument. PS. E9, F63, ..., **PEP**, they define the python protocol. PS. `--select ... --statistics` can be removed. PS. `mypy` 是 linting tool: check the code whether it's correct or not. \
  > (1:42:45)  `pyproject.toml`: the configuration is related to/ visible to our packages. [] 也就是 tag. Inside the tage, you will find out the information. Either you can write it down inside the `setup.cfg` or in the `pyproject.toml`; these two files.

* (1:44:35) 完成 `.github/workflows/ci.yaml`

* (1:45:00) 建立 `.github/workflows/python-publish.yaml`
  > 主要來自於 [Publish Python Package](https://github.com/henrykohl/MongoDB-Connector/actions/new)，點選 **Configure**，就可以看到  Github 編寫好的 yaml/yml 檔案。\
  > `flake8 .` 中那一個點是指 current directory. We are checking into the current directory, whatever thing is wrong like we have not written according to the protocol. It will generate a like warning or information of the message regarding to those thing/those issues. `flake8` 後也可以指定一個 specific folder over here. It means that you can look into the entire workspace. \
  > `pytest`: we are using for testing like we have test cases for that. \ 
  > (2:03:21) this particular configuration for deploying the code on PYPI.

* (1:50:05) review [PyPI · The Python Package Index](https://pypi.org/)
  > 註冊帳號，登入pypi，在 Account setting 中 API tokens，按下 Add API token，在 Create API token 頁面中，設定 Token name (e.g., deconnection) 與 scope (e.g., entire account)，按下 Create token，獲得 TOKEN \
  > 在GitHub 的 setting 中 Secrets and variables，設置 New secret，Name* 為 `PYPI_API_TOKEN`，Secret* 為剛獲取的 TOKEN

* (2:00:15) 解析 `.github/workflows/ci.yaml`
  > This file is not related to PYPI. This is just for the continuous integration. We are going to integrate our code over the GitHub. And GitHub is providing a server. The server name is nothing GitHub Action.GitHub action is a service. Under that we have different servers (e.g., Windows, Linux, ...). In ci.yaml, the particular code/configuration we have written tells us if we are going to integrate our code with GitHub, so immediately we can run all the test cases. We can check our code is fine or not. This ci file is just for testing our code while integrating. This is the work for development.  

* (2:05:20) `src/mongodb_connect/mongo_crud.py`（more details in Lecture 4），`tests/integration/__init__.py`，`tests/__init__.py`，Lecture 3 示範的內容，實際測試時，會發生來自 decorator 的錯誤，但 Lecture 4 示範的內容又不同，卻是可以正確運行的！需要注意。  
  > Everything we have unified in a single method

* (2:08:58) 刪除 `main.py`

* (2:09:15) 執行 git push
  > ```bash
  > git add .
  > git commit -m "everything updated"
  > git push -u origin main
  > ```

* Github Actions 出現錯誤 `from tox.config.cli.parser not found` (Lecture 3 未解決)  

* (2:11:25) GitHub Release
  > 按下 **Create a new release**: \n
  > * Choose a tag: `v0.0.1` \n
  > * Release a title: `0.0.1` \n
  > * Describe a release: `this is my testing package`
  > * 按下 `Publish release`
  >

  > 在 GitHub Actions 頁面可以看到，The package is going to be deployed in backend

* (2:14:14) Github Actions 中 deploy 時，Test with pytest 也出現錯誤 (Lecture 3 未解決) 

---

# Lecture 4 Note -- [GIT & DOCKER For MLOps Projec](https://www.youtube.com/watch?v=KWoyJwqt22I)

## Code Recap (27:00)

* `init_setup.sh`: If you are going to create an environment for that some like repeated command is required like `conda create`、`conda activate`, all which you are running on your bash terminal (on your windows terminal). You can write down each and every command inside your shell script (.sh). And, you can run it in a single shot by using the bash commands.

* (30:13) `/src`: having our main package. It is representing the source code.

* `/experiments`: just for the experiment purpose. Whatever experiment I was doing. Here each and every file I was creating here itself.

so inside this experiment folder, I had created the ipynb file.

* (30:45) `/.github/workflows`: inside the workflows folder, you can find out the complete configuration. The configuration is related to the Continuous Integration and The configuration is related to the PYPI deployment. When we are going to deploy our code on a PYPI repository, for that we have written a configuratin inside this particular yaml file (`python-publish.yaml`). For the continuous integration, if I want to test our code on GitHub Action server, so for that we have written `ci.yaml` file as soon as we want to integrate our code or as soon as we are going to push our code to the GitHub. So the Github Action server will up and then like each and every script is going to be run.

* (32:51) `experiments/experiments.ipynb` 
  > (44:10) MongoDB --> client --> Database --> Collection <--> Document:{key, value}. Inside the collection my data will be available in the form of document. document is nothing key and value pair.

* (37:00) MongoDB

* (42:10-1:35:25) 完成 `experiments/experiments.ipynb` 

* (1:35:25) Review `src/mongodb_connect/mongo_crud.py`
  > 注意，`mongo_crud.py` 要使用 [MongoDB-Connector/src/mongodb_connect
/mongo_crud.py](https://github.com/sunnysavita10/MongoDB-Connector/blob/main/src/mongodb_connect/mongo_crud.py)，而不是使用[MongoDB-Connector-PYPI-Package/src/database_automation
/mongo_crud.py](https://github.com/sunnysavita10/MongoDB-Connector-PYPI-Package/blob/main/src/database_automation/mongo_crud.py)

* (1:36:20) 將 `setup.py` 的 PKG_NAME 改為 "databaseautomation"， 將 `/src` 下的資料夾 改名為 *database_automation*. PS. __version__ = 0.0.2 改成 0.0.3 ??
  > ```bash
  > git add .
  > git commit -m "code and version updated"
  > git push -u origin main
  > ```

* In Github Action webpage，the code is running on top of github action. 

* 在 Github repository 的 release 頁面，可以看到 v0.0.3 的版本

* (1:44:14) 將 `setup.py` 的 __version__ = 0.0.4
  > ``` bash
  > git add .
  > git commit -m "version updated"
  > git push origin main
  > ```

* 在 [github releases 頁面](https://github.com/henrykohl/MongoDB-Connector/releases) ，點選 **Draft a new release**，在 **Choose a Tag** 中選取 v0.0.4，在 release title 輸入 v0.0.4，在 Describe release 中輸入 this is my mongo automation package. 最後按下 **Publish release**.

* (1:48:00) Google Colab，自行將 Lecture 的演示，記錄到 `mongodemo.ipynb`

* (2:10:54) Explanation of YAML file

---

# 參考

* 參1: [MongoDB Basics | Tutorial 4: Create Atlas Cluster](https://www.youtube.com/watch?v=esKNjzDZItQ)

* 參2: [Simple documentation for setup.cfg](https://discuss.python.org/t/simple-documentation-for-setup-cfg/11465)

---
# requirements_dev.txt we use for the testing
It makes it easier to install and manage dependencies for development and testing, separate from the dependencies required for production.

# difference between requirements_dev.txt and requirements.txt

requirements.txt is used to specify the dependencies required to run the production code of a Python project, while requirements_dev.txt is used to specify the dependencies required for development and testing purposes.

# tox.ini
We use if for the testing in the python package testing against different version of the python 

## how tox works tox enviornment creation
1. Install depedencies and packages 
2. Run commands
3. Its a combination of the (virtualenvwrapper and makefile)
4. It creates a .tox


# pyproject.toml
it is being used for configuration the python project it is a alternative of the setup.cfg file. its containts configuration related to the build system
such as the build tool used package name version author license and dependencies

# setup.cfg
In summary, setup.cfg is used by setuptools to configure the packaging and installation of a Python projec

# Testing python application
*types of testing*
1. Automated testing 
2. Manual testing

*Mode of testing*
1. Unit testing
2. Integration tests

*Testing frameworks*

1. pytest
2. unittest
3. robotframework
4. selenium
5. behave
6. doctest

# check with the code style formatting and syntax(coding standard)

1. pylint
2. flake8(it is best because it containt 3 library pylint pycodestyle mccabe)
3. pycodestyle


